/*
  IeeFlip Library - Declaration file
  
  Originally created 5 March 2011

  This file is in the public domain.  
*/

#ifndef IeeFlip_h
#define IeeFlip_h

#include <inttypes.h>
#include "Print.h"
#include "WProgram.h"

#define IEE_FLIP_PERFORM_LONG_TIMING_DELAYS 1 // Don't seem to need the long timing delays. If the display is dropping characters, try turning this on.
#define IEE_FLIP_OLD_SCHOOL_ALPHANUMERIC_DISPLAY 1 // Turning this off has not been tested as a newer dot matrix display was not available.

class IeeFlip : public Print {
public:

  // Special control characters supported by IEE-Flip displays.
  // Send using controlCommand(), not write().
  // Few devices support all control characters.
#if IEE_FLIP_OLD_SCHOOL_ALPHANUMERIC_DISPLAY

  // These were found on a Model 03700-06-020/S 1x20 alphanumeric display by experimentation.
  static const uint8_t BACKSPACE_CHARACTER = 0x08; 
  static const uint8_t ADVANCE_CURSOR__CHARACTER = 0x09; // Tab forward 1 character.
  static const uint8_t LINE_FEED_CHARACTER = 0x0A;
  static const uint8_t BEGIN_BLINKING_CHARS_CHARACTER = 0x0B; 
  static const uint8_t END_BLINKING_CHARS_CHARACTER = 0x0C; 
  static const uint8_t CARRIAGE_RETURN_CHARACTER = 0x0D;
  static const uint8_t BLINK_CURSOR_OFF_CHARACTER = 0x0E; 
  static const uint8_t BLINK_CURSOR_ON_CHARACTER = 0x0F; 

  static const uint8_t END_OF_LINE_WRAP_CHARATCER = 0x11;
  static const uint8_t END_OF_LINE_NO_ADVANCE_CHARATCER = 0x12; // Cursor stops on the last position at end of display.
  static const uint8_t END_OF_LINE_HORIZONTAL_SCROLL_CHARATCER = 0x13;
  static const uint8_t CURSOR_SET_CHARACTER = 0x1B; // Second character is cursor address.
  
#else // IEE_FLIP_OLD_SCHOOL_ALPHANUMERIC_DISPLAY

  static const uint8_t READ_DISPLAY_ID_CHARACTER = 0x01; 
  static const uint8_t READ_SOFTWARE_CHECKSUM_CHARACTER = 0x02; 
  static const uint8_t READ_CURSOR_LOCATION_CHARACTER = 0x03; 
  static const uint8_t READ_DATA_AT_CURSOR_CHARACTER = 0x04; 
  static const uint8_t READ_DATA_AT_CURSOR_AND_INCREMENT_CHARACTER = 0x05; 
  static const uint8_t BELL_ALARM_OUTPUT_CHARACTER = 0x07; 
  static const uint8_t BACKSPACE_CHARACTER = 0x08; 
  static const uint8_t ADVANCE_CURSOR__CHARACTER = 0x09; // Tab forward 1 character.
  static const uint8_t LINE_FEED_CHARACTER = 0x0A;
  static const uint8_t BLINK_BLOCK_CURSOR_CHARACTER = 0x0B; 
  static const uint8_t UNDERBAR_CURSOR_CHARACTER = 0x0C; 
  static const uint8_t CARRIAGE_RETURN_CHARACTER = 0x0D; 
  static const uint8_t CURSOR_OFF_CHARACTER = 0x0E; 
  static const uint8_t CURSOR_ON_CHARACTER = 0x0F; 
  static const uint8_t SCROLL_LINE_LOCK_CHARACTER = 0x10; 
  static const uint8_t VERICAL_SCROLL_MODE_CHARACTER = 0x11; 
  static const uint8_t HORIZONTAL_SCROLL_MODE_CHARACTER = 0x13; 
  static const uint8_t SOFTWARE_RESET_CHARACTER = 0x14; 
  static const uint8_t CLEAR_DISPLAY_HOME_CURSOR_CHARACTER = 0x15; 
  static const uint8_t HOME_CURSOR_CHARACTER = 0x16; 
  static const uint8_t SET_DATA_BIT_7_HIGH_CHARACTER = 0x17; // Data bit 7 is high for next character. 
  static const uint8_t BEGIN_USER_DEFINED_CHAR_CHARACTER = 0x18; // Sennd location, followed by data bytes. 
  static const uint8_t SET_ADDRESS_BIT_0_HIGH_CHARACTER = 0x19; // Address bit 0 high for next character. 
  static const uint8_t CURSOR_UP_LINE_CHARACTER = 0x1A; 
  static const uint8_t CURSOR_SET_CHARACTER = 0x1B; // Second character is cursor address.
  static const uint8_t SELECT_EUROPEAN_CHAR_SET_CHARACTER = 0x1C; 
  static const uint8_t SELECT_KATAKANA_CHAR_SET_CHARACTER = 0x1D; 
  static const uint8_t SELECT_CYRILLIC_CHAR_SET_CHARACTER = 0x1E; 
  static const uint8_t SELECT_HEBREW_CHAR_SET_CHARACTER = 0x1F;

  // These commands require address0 to be connected.  
  static const uint8_t SET_BRIGHTNESS_CHARACTER = 0x30; // Column and brightness bytes follow. Column 0xFF = all columns. 
  static const uint8_t BEGIN_BLINKING_CHARS_CHARACTER = 0x31; 
  static const uint8_t END_BLINKING_CHARS_CHARACTER = 0x32; 
  static const uint8_t BLANK_DISPLAY_CHARACTER = 0x33; 
  static const uint8_t UNBLANK_DISPLAY_CHARACTER = 0x34; 
  static const uint8_t COMMA_PERIOD_TRIANGLE_FUNCTION_CHARACTER = 0x35; 
  static const uint8_t ERASE_LINE_WITH_END_BLINK_CHARACTER = 0x36; 
  static const uint8_t SET_CARRIAGE_RETURN_LINE_FEED_DEFINITIONS_CHARACTER = 0x37; 
  static const uint8_t UNDERBAR_ON_CHARACTER = 0x39; 
  static const uint8_t UNDERBAR_OFF_CHARACTER = 0x39; 
  static const uint8_t RIGHT_TO_LEFT_CHARACTER = 0x3A; 
  static const uint8_t LEFT_TO_RIGHT_CHARACTER = 0x3B; 
  static const uint8_t SCREEN_SAVER_ON_CHARACTER = 0x3C; 
  static const uint8_t SCREEN_SAVER_OFF_CHARACTER = 0x3D; 
  static const uint8_t SELF_TEST_CHARACTER = 0x3E; 
  static const uint8_t TERMINATE_SELF_TEST_CHARACTER = 0x3F; 

#endif // IEE_FLIP_OLD_SCHOOL_ALPHANUMERIC_DISPLAY
	
  // May be used for busyPin and data7Pin.
  static const uint8_t DO_NOT_USE_PIN = 0xFF;

  // This library supports multiple displays. In this case, all pins
  // can be shared except chip select. This library does not support
  // sharing pins with other devices. Arduino pin 13 (the one connected
  // to the on board LED) can be used for data lines but it can not
  // be used for any of the control lines. 
  // To save pins, rdPin, address0Pin and/or data7Pin can be set to
  // IeeFlip::DO_NOT_USE_PIN. When using IeeFlip::DO_NOT_USE_PIN, /RD 
  // should be left unconnected, address0Pin should be tied to ground, 
  // data7Pin should be tied to ground with a 10k resistor.
  // TODO: Test multiple displays.
  IeeFlip(uint8_t csPin,
          uint8_t wrPin,
          uint8_t rdPin,
          uint8_t address0Pin,
          uint8_t data0Pin, uint8_t data1Pin, uint8_t data2Pin, uint8_t data3Pin,
          uint8_t data4Pin, uint8_t data5Pin, uint8_t data6Pin, uint8_t data7Pin);
             
  // Sets the character dimensions of the display.
  void begin(uint8_t numColumns, int numRows);

  void clear();
  void home();

#if !IEE_FLIP_OLD_SCHOOL_ALPHANUMERIC_DISPLAY // Doesn't seem to be supported on old displays.
  void noDisplay();
  void display();
#endif
  void noBlink();
  void blink();
  void noCursor(); // On old displays, this is the same as blink.
  void cursor();
  
  // Methods supported by LiquidCrystal but not by our library.
//  void scrollDisplayLeft();
//  void scrollDisplayRight();
//  void leftToRight();
//  void rightToLeft();
//  void autoscroll();
//  void noAutoscroll();
//  void createChar(uint8_t, uint8_t[]);
//  void command(uint8_t);

  void setCursor(uint8_t column, uint8_t row); 
  virtual void write(uint8_t character);

  // Methods unique to our class.
  void controlCharacter(uint8_t character); // Commands are different from HD44780/LiquidCrystal.

#if !IEE_FLIP_OLD_SCHOOL_ALPHANUMERIC_DISPLAY // Can't find any read commands that work with the old display.
  uint8_t read(uint8_t address0Level);
#endif

private:

  static const uint8_t NUM_DATA_PINS = 8;
  static const uint8_t FIRST_HIGH_CONTROL_CHARACTER = 0x20;

  // These wait times seem exceedingly long. Came from IEE S036X2 datasheet.
  // Timings vary depending on the display, the longest in each case is used below.
  // The display tested did misbehave when there was no delay added at all.
  static const uint16_t SEND_CHARACTER_WAIT_TIME_IN_MICROSECONDS = 460;
  static const uint16_t BLINK_SET_WAIT_TIME_IN_MICROSECONDS = 470 - SEND_CHARACTER_WAIT_TIME_IN_MICROSECONDS;
  static const uint16_t BRIGHTNESS_SET_WAIT_TIME_IN_MICROSECONDS = 465 - SEND_CHARACTER_WAIT_TIME_IN_MICROSECONDS;
  static const uint16_t CLEAR_WAIT_TIME_IN_MICROSECONDS = 1726 - SEND_CHARACTER_WAIT_TIME_IN_MICROSECONDS;
  static const uint16_t CURSOR_SET_WAIT_TIME_IN_MICROSECONDS = 460 - SEND_CHARACTER_WAIT_TIME_IN_MICROSECONDS;
  static const uint16_t HOME_WAIT_TIME_IN_MICROSECONDS = 460 - SEND_CHARACTER_WAIT_TIME_IN_MICROSECONDS;
  static const uint16_t HSCROLL_WAIT_TIME_IN_MICROSECONDS = 2738 - SEND_CHARACTER_WAIT_TIME_IN_MICROSECONDS;
  static const uint16_t RESET_WAIT_TIME_IN_MICROSECONDS = 1000 - SEND_CHARACTER_WAIT_TIME_IN_MICROSECONDS;
  static const uint16_t VSCROLL_WAIT_TIME_IN_MICROSECONDS = 3504 - SEND_CHARACTER_WAIT_TIME_IN_MICROSECONDS;
  static const uint16_t READ_CHARACTER_WAIT_TIME_IN_MICROSECONDS = 460; // Not documented, guess same as a write.
  
  uint8_t _csPin;
  uint8_t _wrPin;
  uint8_t _rdPin;
  uint8_t _address0Pin;
  uint8_t _dataPins[NUM_DATA_PINS];
  
  uint8_t _numDisplayColumns;
  uint8_t _numDisplayRows;
  uint8_t _numDisplayCharacters;

  void writeInternal(uint8_t character, uint8_t address0Level);
  inline void timingDelay(uint16_t delayInMicroseconds);
};

inline void IeeFlip::timingDelay(uint16_t delayInMicroseconds)
{
#if IEE_FLIP_PERFORM_LONG_TIMING_DELAYS
  delayMicroseconds(delayInMicroseconds);
#endif
}
  
#endif // IeeFlip_h

